import textwrap
from typing import Sequence

from ai.backend.client.output.fields import storage_fields
from ai.backend.client.output.types import FieldSpec, PaginatedResult
from ai.backend.client.pagination import generate_paginated_results
from ai.backend.client.session import api_session

from .base import BaseFunction, api_function

__all__ = ("Storage",)

_default_list_fields = (
    storage_fields["id"],
    storage_fields["backend"],
    storage_fields["capabilities"],
)

_default_detail_fields = (
    storage_fields["id"],
    storage_fields["backend"],
    storage_fields["path"],
    storage_fields["fsprefix"],
    storage_fields["capabilities"],
    storage_fields["hardware_metadata"],
)


class Storage(BaseFunction):
    """
    Provides a shortcut of :func:`Admin.query()
    <ai.backend.client.admin.Admin.query>` that fetches various straoge volume
    information keyed by vfolder hosts.

    .. note::

      All methods in this function class require your API access key to
      have the *super-admin* privilege.
    """

    @api_function
    @classmethod
    async def paginated_list(
        cls,
        status: str = "ALIVE",
        *,
        fields: Sequence[FieldSpec] = _default_list_fields,
        page_offset: int = 0,
        page_size: int = 20,
        filter: str = None,
        order: str = None,
    ) -> PaginatedResult[dict]:
        """
        Lists the keypairs.
        You need an admin privilege for this operation.
        """
        return await generate_paginated_results(
            "storage_volume_list",
            {
                "filter": (filter, "String"),
                "order": (order, "String"),
            },
            fields,
            page_offset=page_offset,
            page_size=page_size,
        )

    @api_function
    @classmethod
    async def detail(
        cls,
        vfolder_host: str,
        fields: Sequence[FieldSpec] = _default_detail_fields,
    ) -> dict:
        query = textwrap.dedent(
            """\
            query($vfolder_host: String!) {
                storage_volume(id: $vfolder_host) {$fields}
            }
        """
        )
        query = query.replace("$fields", " ".join(f.field_ref for f in fields))
        variables = {"vfolder_host": vfolder_host}
        data = await api_session.get().Admin._query(query, variables)
        return data["storage_volume"]



from __future__ import annotations

import sys

import click
import humanize
from tabulate import tabulate

from ai.backend.client.func.vfolder import _default_list_fields
from ai.backend.client.session import Session

from ..extensions import pass_ctx_obj
from ..pretty import print_error
from ..types import CLIContext
from ..vfolder import vfolder as user_vfolder
from . import admin


@admin.group()
def vfolder() -> None:
    """
    VFolder administration commands.
    """


def _list_cmd(docs: str = None):
    @pass_ctx_obj
    @click.option("-g", "--group", type=str, default=None, help="Filter by group ID.")
    @click.option("--filter", "filter_", default=None, help="Set the query filter expression.")
    @click.option("--order", default=None, help="Set the query ordering expression.")
    @click.option("--offset", default=0, help="The index of the current page start for pagination.")
    @click.option("--limit", default=None, help="The page size for pagination.")
    def list(ctx: CLIContext, group, filter_, order, offset, limit) -> None:
        """
        List virtual folders.
        """
        try:
            with Session() as session:
                fetch_func = lambda pg_offset, pg_size: session.VFolder.paginated_list(
                    group,
                    fields=_default_list_fields,
                    page_offset=pg_offset,
                    page_size=pg_size,
                    filter=filter_,
                    order=order,
                )
                ctx.output.print_paginated_list(
                    fetch_func,
                    initial_page_offset=offset,
                    page_size=limit,
                )
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(1)

    if docs is not None:
        list.__doc__ = docs
    return list


user_vfolder.command()(_list_cmd())
vfolder.command()(_list_cmd())


@vfolder.command()
def list_hosts():
    """
    List all mounted hosts from virtual folder root.
    (superadmin privilege required)
    """
    with Session() as session:
        try:
            resp = session.VFolder.list_all_hosts()
            print("Default vfolder host: {}".format(resp["default"]))
            print("Mounted hosts: {}".format(", ".join(resp["allowed"])))
        except Exception as e:
            print_error(e)
            sys.exit(1)


@vfolder.command()
@click.argument("vfolder_host")
def perf_metric(vfolder_host):
    """
    Show the performance statistics of a vfolder host.
    (superadmin privilege required)

    A vfolder host consists of a string of the storage proxy name and the volume name
    separated by a colon. (e.g., "local:volume1")
    """
    with Session() as session:
        try:
            resp = session.VFolder.get_performance_metric(vfolder_host)
            print(
                tabulate(
                    [
                        (k, humanize.naturalsize(v, binary=True) if "bytes" in k else f"{v:.2f}")
                        for k, v in resp["metric"].items()
                    ],
                    headers=("Key", "Value"),
                )
            )
        except Exception as e:
            print_error(e)
            sys.exit(1)


@vfolder.command()
@click.option(
    "-a", "--agent-id", type=str, default=None, help="Target agent to fetch fstab contents."
)
def get_fstab_contents(agent_id):
    """
    Get contents of fstab file from a node.
    (superadmin privilege required)

    If agent-id is not specified, manager's fstab contents will be returned.
    """
    with Session() as session:
        try:
            resp = session.VFolder.get_fstab_contents(agent_id)
        except Exception as e:
            print_error(e)
            sys.exit(1)
        print(resp)


@vfolder.command()
def list_mounts():
    """
    List all mounted hosts in virtual folder root.
    (superadmin privilege required)
    """
    with Session() as session:
        try:
            resp = session.VFolder.list_mounts()
        except Exception as e:
            print_error(e)
            sys.exit(1)
        print("manager")
        for k, v in resp["manager"].items():
            print(" ", k, ":", v)
        print("\nagents")
        for aid, data in resp["agents"].items():
            print(" ", aid)
            for k, v in data.items():
                print("   ", k, ":", v)


@vfolder.command()
@click.argument("fs-location", type=str)
@click.argument("name", type=str)
@click.option("-o", "--options", type=str, default=None, help="Mount options.")
@click.option("--edit-fstab", is_flag=True, help="Edit fstab file to mount permanently.")
def mount_host(fs_location, name, options, edit_fstab):
    """
    Mount a host in virtual folder root.
    (superadmin privilege required)

    \b
    FS-LOCATION: Location of file system to be mounted.
    NAME: Name of mounted host.
    """
    with Session() as session:
        try:
            resp = session.VFolder.mount_host(name, fs_location, options, edit_fstab)
        except Exception as e:
            print_error(e)
            sys.exit(1)
        print("manager")
        for k, v in resp["manager"].items():
            print(" ", k, ":", v)
        print("agents")
        for aid, data in resp["agents"].items():
            print(" ", aid)
            for k, v in data.items():
                print("   ", k, ":", v)


@vfolder.command()
@click.argument("name", type=str)
@click.option("--edit-fstab", is_flag=True, help="Edit fstab file to mount permanently.")
def umount_host(name, edit_fstab):
    """
    Unmount a host from virtual folder root.
    (superadmin privilege required)

    \b
    NAME: Name of mounted host.
    """
    with Session() as session:
        try:
            resp = session.VFolder.umount_host(name, edit_fstab)
        except Exception as e:
            print_error(e)
            sys.exit(1)
        print("manager")
        for k, v in resp["manager"].items():
            print(" ", k, ":", v)
        print("agents")
        for aid, data in resp["agents"].items():
            print(" ", aid)
            for k, v in data.items():
                print("   ", k, ":", v)



import sys

import click

from ai.backend.cli.main import main

from ..compat import asyncio_run
from ..session import AsyncSession
from .pretty import print_error


@main.command()
@click.argument("task_id", metavar="TASKID")
def task_logs(task_id):
    """
    Shows the output logs of a batch task.

    \b
    TASKID: An UUID of a task (or kernel).
    """

    async def _task_logs():
        async with AsyncSession() as session:
            async for chunk in session.ComputeSession.get_task_logs(task_id):
                print(chunk.decode("utf8", errors="replace"), end="")

    try:
        asyncio_run(_task_logs())
    except Exception as e:
        print_error(e)
        sys.exit(1)



from __future__ import annotations

import textwrap
from typing import Sequence

from ai.backend.client.output.fields import agent_fields
from ai.backend.client.output.types import FieldSpec, PaginatedResult
from ai.backend.client.pagination import generate_paginated_results
from ai.backend.client.request import Request
from ai.backend.client.session import api_session

from .base import BaseFunction, api_function

__all__ = (
    "Agent",
    "AgentWatcher",
)

_default_list_fields = (
    agent_fields["id"],
    agent_fields["status"],
    agent_fields["scaling_group"],
    agent_fields["available_slots"],
    agent_fields["occupied_slots"],
)

_default_detail_fields = (
    agent_fields["id"],
    agent_fields["status"],
    agent_fields["scaling_group"],
    agent_fields["addr"],
    agent_fields["region"],
    agent_fields["first_contact"],
    agent_fields["cpu_cur_pct"],
    agent_fields["mem_cur_bytes"],
    agent_fields["available_slots"],
    agent_fields["occupied_slots"],
)


class Agent(BaseFunction):
    """
    Provides a shortcut of :func:`Admin.query()
    <ai.backend.client.admin.Admin.query>` that fetches various agent
    information.

    .. note::

      All methods in this function class require your API access key to
      have the *admin* privilege.
    """

    @api_function
    @classmethod
    async def paginated_list(
        cls,
        status: str = "ALIVE",
        scaling_group: str = None,
        *,
        fields: Sequence[FieldSpec] = _default_list_fields,
        page_offset: int = 0,
        page_size: int = 20,
        filter: str = None,
        order: str = None,
    ) -> PaginatedResult:
        """
        Lists the keypairs.
        You need an admin privilege for this operation.
        """
        return await generate_paginated_results(
            "agent_list",
            {
                "status": (status, "String"),
                "scaling_group": (scaling_group, "String"),
                "filter": (filter, "String"),
                "order": (order, "String"),
            },
            fields,
            page_size=page_size,
            page_offset=page_offset,
        )

    @api_function
    @classmethod
    async def detail(
        cls,
        agent_id: str,
        fields: Sequence[FieldSpec] = _default_detail_fields,
    ) -> Sequence[dict]:
        query = textwrap.dedent(
            """\
            query($agent_id: String!) {
                agent(agent_id: $agent_id) {$fields}
            }
        """
        )
        query = query.replace("$fields", " ".join(f.field_ref for f in fields))
        variables = {"agent_id": agent_id}
        data = await api_session.get().Admin._query(query, variables)
        return data["agent"]


class AgentWatcher(BaseFunction):
    """
    Provides a shortcut of :func:`Admin.query()
    <ai.backend.client.admin.Admin.query>` that manipulate agent status.

    .. note::

      All methods in this function class require you to
      have the *superadmin* privilege.
    """

    @api_function
    @classmethod
    async def get_status(cls, agent_id: str) -> dict:
        """
        Get agent and watcher status.
        """
        rqst = Request("GET", "/resource/watcher")
        rqst.set_json({"agent_id": agent_id})
        async with rqst.fetch() as resp:
            data = await resp.json()
            if "message" in data:
                return data["message"]
            else:
                return data

    @api_function
    @classmethod
    async def agent_start(cls, agent_id: str) -> dict:
        """
        Start agent.
        """
        rqst = Request("POST", "/resource/watcher/agent/start")
        rqst.set_json({"agent_id": agent_id})
        async with rqst.fetch() as resp:
            data = await resp.json()
            if "message" in data:
                return data["message"]
            else:
                return data

    @api_function
    @classmethod
    async def agent_stop(cls, agent_id: str) -> dict:
        """
        Stop agent.
        """
        rqst = Request("POST", "/resource/watcher/agent/stop")
        rqst.set_json({"agent_id": agent_id})
        async with rqst.fetch() as resp:
            data = await resp.json()
            if "message" in data:
                return data["message"]
            else:
                return data

    @api_function
    @classmethod
    async def agent_restart(cls, agent_id: str) -> dict:
        """
        Restart agent.
        """
        rqst = Request("POST", "/resource/watcher/agent/restart")
        rqst.set_json({"agent_id": agent_id})
        async with rqst.fetch() as resp:
            data = await resp.json()
            if "message" in data:
                return data["message"]
            else:
                return data



"""
A compatibility module for backported codes from Python 3.6+ standard library.
"""

import asyncio

__all__ = (
    "current_loop",
    "all_tasks",
    "asyncio_run",
    "asyncio_run_forever",
)


if hasattr(asyncio, "get_running_loop"):  # Python 3.7+
    current_loop = asyncio.get_running_loop
else:
    current_loop = asyncio.get_event_loop


if hasattr(asyncio, "all_tasks"):  # Python 3.7+
    all_tasks = asyncio.all_tasks
else:
    all_tasks = asyncio.Task.all_tasks  # type: ignore


def _cancel_all_tasks(loop):
    to_cancel = all_tasks(loop)
    if not to_cancel:
        return
    for task in to_cancel:
        task.cancel()
    loop.run_until_complete(asyncio.gather(*to_cancel, return_exceptions=True))
    for task in to_cancel:
        if task.cancelled():
            continue
        if task.exception() is not None:
            loop.call_exception_handler(
                {
                    "message": "unhandled exception during asyncio.run() shutdown",
                    "exception": task.exception(),
                    "task": task,
                }
            )


def _asyncio_run(coro, *, debug=False):
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    loop.set_debug(debug)
    try:
        return loop.run_until_complete(coro)
    finally:
        try:
            _cancel_all_tasks(loop)
            if hasattr(loop, "shutdown_asyncgens"):  # Python 3.6+
                loop.run_until_complete(loop.shutdown_asyncgens())
        finally:
            loop.stop()
            loop.close()
            asyncio.set_event_loop(None)


if hasattr(asyncio, "run"):  # Python 3.7+
    asyncio_run = asyncio.run
else:
    asyncio_run = _asyncio_run


def asyncio_run_forever(server_context, *, debug=False):
    """
    A proposed-but-not-implemented asyncio.run_forever() API based on
    @vxgmichel's idea.
    See discussions on https://github.com/python/asyncio/pull/465
    """
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    loop.set_debug(debug)

    forever = loop.create_future()

    async def _run_forever():
        async with server_context:
            try:
                await forever
            except asyncio.CancelledError:
                pass

    try:
        return loop.run_until_complete(_run_forever())
    finally:
        try:
            _cancel_all_tasks(loop)
            if hasattr(loop, "shutdown_asyncgens"):  # Python 3.6+
                loop.run_until_complete(loop.shutdown_asyncgens())
        finally:
            loop.stop()
            loop.close()
            asyncio.set_event_loop(None)



from typing import Mapping

from ..request import Request
from .base import BaseFunction, api_function

__all__ = ("System",)


class System(BaseFunction):
    """
    Provides the function interface for the API endpoint's system information.
    """

    @api_function
    @classmethod
    async def get_versions(cls) -> Mapping[str, str]:
        rqst = Request("GET", "/")
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def get_manager_version(cls) -> str:
        rqst = Request("GET", "/")
        async with rqst.fetch() as resp:
            ret = await resp.json()
            return ret["manager"]

    @api_function
    @classmethod
    async def get_api_version(cls) -> str:
        rqst = Request("GET", "/")
        async with rqst.fetch() as resp:
            ret = await resp.json()
            return ret["version"]



import asyncio
import json
import shlex
import sys
from typing import Dict, List, MutableMapping, Optional, Sequence, Union

import aiohttp
import click

from ai.backend.cli.main import main

from ..compat import asyncio_run, asyncio_run_forever
from ..config import DEFAULT_CHUNK_SIZE
from ..request import Request
from ..session import AsyncSession
from ..versioning import get_naming
from .pretty import print_error, print_fail, print_info, print_warn


class WSProxy:
    __slots__ = (
        "api_session",
        "session_name",
        "app_name",
        "args",
        "envs",
        "reader",
        "writer",
    )

    def __init__(
        self,
        api_session: AsyncSession,
        session_name: str,
        app_name: str,
        args: MutableMapping[str, Union[None, str, List[str]]],
        envs: MutableMapping[str, str],
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
    ) -> None:
        self.api_session = api_session
        self.session_name = session_name
        self.app_name = app_name
        self.args = args
        self.envs = envs
        self.reader = reader
        self.writer = writer

    async def run(self) -> None:
        prefix = get_naming(self.api_session.api_version, "path")
        path = f"/stream/{prefix}/{self.session_name}/tcpproxy"
        params = {"app": self.app_name}

        if len(self.args.keys()) > 0:
            params["arguments"] = json.dumps(self.args)
        if len(self.envs.keys()) > 0:
            params["envs"] = json.dumps(self.envs)

        api_rqst = Request("GET", path, b"", params=params, content_type="application/json")
        async with api_rqst.connect_websocket() as ws:

            async def downstream() -> None:
                try:
                    async for msg in ws:
                        if msg.type == aiohttp.WSMsgType.ERROR:
                            await self.write_error(msg)
                            break
                        elif msg.type == aiohttp.WSMsgType.CLOSE:
                            if msg.data != aiohttp.WSCloseCode.OK:
                                await self.write_error(msg)
                            break
                        elif msg.type == aiohttp.WSMsgType.BINARY:
                            self.writer.write(msg.data)
                            await self.writer.drain()
                except ConnectionResetError:
                    pass  # shutting down
                except asyncio.CancelledError:
                    pass
                finally:
                    self.writer.close()
                    try:
                        await self.writer.wait_closed()
                    except (BrokenPipeError, IOError):
                        # closed
                        pass

            down_task = asyncio.create_task(downstream())
            try:
                while True:
                    chunk = await self.reader.read(DEFAULT_CHUNK_SIZE)
                    if not chunk:
                        break
                    await ws.send_bytes(chunk)
            except ConnectionResetError:
                pass  # shutting down
            except asyncio.CancelledError:
                raise
            finally:
                if not down_task.done():
                    down_task.cancel()
                    await down_task

    async def write_error(self, msg: aiohttp.WSMessage) -> None:
        if isinstance(msg.data, bytes):
            error_msg = msg.data.decode("utf8")
        else:
            error_msg = str(msg.data)
        rsp = (
            "HTTP/1.1 503 Service Unavailable\r\n"
            "Connection: Closed\r\n\r\n"
            "WebSocket reply: {}".format(error_msg)
        )
        self.writer.write(rsp.encode())
        await self.writer.drain()


class ProxyRunnerContext:

    __slots__ = (
        "session_name",
        "app_name",
        "protocol",
        "host",
        "port",
        "args",
        "envs",
        "api_session",
        "local_server",
        "exit_code",
    )

    session_name: str
    app_name: str
    protocol: str
    host: str
    port: int
    args: Dict[str, Union[None, str, List[str]]]
    envs: Dict[str, str]
    api_session: Optional[AsyncSession]
    local_server: Optional[asyncio.AbstractServer]
    exit_code: int

    def __init__(
        self,
        host: str,
        port: int,
        session_name: str,
        app_name: str,
        *,
        protocol: str = "tcp",
        args: Sequence[str] = None,
        envs: Sequence[str] = None,
    ) -> None:
        self.host = host
        self.port = port
        self.session_name = session_name
        self.app_name = app_name
        self.protocol = protocol

        self.api_session = None
        self.local_server = None
        self.exit_code = 0

        self.args, self.envs = {}, {}
        if args is not None and len(args) > 0:
            for argline in args:
                tokens = []
                for token in shlex.shlex(argline, punctuation_chars=True):
                    kv = token.split("=", maxsplit=1)
                    if len(kv) == 1:
                        tokens.append(shlex.split(token)[0])
                    else:
                        tokens.append(kv[0])
                        tokens.append(shlex.split(kv[1])[0])

                if len(tokens) == 1:
                    self.args[tokens[0]] = None
                elif len(tokens) == 2:
                    self.args[tokens[0]] = tokens[1]
                else:
                    self.args[tokens[0]] = tokens[1:]
        if envs is not None and len(envs) > 0:
            for envline in envs:
                split = envline.strip().split("=", maxsplit=2)
                if len(split) == 2:
                    self.envs[split[0]] = split[1]
                else:
                    self.envs[split[0]] = ""

    async def handle_connection(
        self,
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
    ) -> None:
        assert self.api_session is not None
        p = WSProxy(
            self.api_session,
            self.session_name,
            self.app_name,
            self.args,
            self.envs,
            reader,
            writer,
        )
        try:
            await p.run()
        except asyncio.CancelledError:
            pass
        except Exception as e:
            print_error(e)

    async def __aenter__(self) -> None:
        self.exit_code = 0
        self.api_session = AsyncSession()
        await self.api_session.__aenter__()

        user_url_template = "{protocol}://{host}:{port}"
        try:
            compute_session = self.api_session.ComputeSession(self.session_name)
            all_apps = await compute_session.stream_app_info()
            for app_info in all_apps:
                if app_info["name"] == self.app_name:
                    if "url_template" in app_info.keys():
                        user_url_template = app_info["url_template"]
                    break
            else:
                print_fail(f'The app "{self.app_name}" is not supported by the session.')
                self.exit_code = 1
                return

            self.local_server = await asyncio.start_server(
                self.handle_connection, self.host, self.port
            )
            user_url = user_url_template.format(
                protocol=self.protocol,
                host=self.host,
                port=self.port,
            )
            print_info(
                'A local proxy to the application "{0}" '.format(self.app_name)
                + 'provided by the session "{0}" '.format(self.session_name)
                + "is available at:\n{0}".format(user_url),
            )
            if self.host == "0.0.0.0":
                print_warn(
                    'NOTE: Replace "0.0.0.0" with the actual hostname you use '
                    "to connect with the CLI app proxy."
                )
        except Exception:
            await self.api_session.__aexit__(*sys.exc_info())
            raise

    async def __aexit__(self, *exc_info) -> None:
        if self.local_server is not None:
            print_info("Shutting down....")
            self.local_server.close()
            await self.local_server.wait_closed()
        assert self.api_session is not None
        await self.api_session.__aexit__(*exc_info)
        assert self.api_session.closed
        if self.local_server is not None:
            print_info('The local proxy to "{}" has terminated.'.format(self.app_name))
        self.local_server = None


@main.command()
@click.argument("session_name", type=str, metavar="NAME")
@click.argument("app", type=str)
@click.option(
    "-b",
    "--bind",
    type=str,
    default="127.0.0.1:8080",
    metavar="[HOST:]PORT",
    help="The IP/host address and the port number to bind this proxy.",
)
@click.option(
    "--arg",
    type=str,
    multiple=True,
    metavar='"--option <value>"',
    help="Add additional argument when starting service.",
)
@click.option(
    "-e",
    "--env",
    type=str,
    multiple=True,
    metavar='"ENVNAME=envvalue"',
    help="Add additional environment variable when starting service.",
)
def app(session_name, app, bind, arg, env):
    """
    Run a local proxy to a service provided by Backend.AI compute sessions.

    The type of proxy depends on the app definition: plain TCP or HTTP.

    \b
    SESSID: The compute session ID.
    APP: The name of service provided by the given session.
    """
    bind_parts = bind.rsplit(":", maxsplit=1)
    if len(bind_parts) == 1:
        host = "127.0.0.1"
        port = int(bind_parts[0])
    elif len(bind_parts) == 2:
        host = bind_parts[0]
        port = int(bind_parts[1])
    try:
        proxy_ctx = ProxyRunnerContext(
            host,
            port,
            session_name,
            app,
            protocol="tcp",
            args=arg,
            envs=env,
        )
        asyncio_run_forever(proxy_ctx)
        sys.exit(proxy_ctx.exit_code)
    except Exception as e:
        print_error(e)
        sys.exit(1)


@main.command()
@click.argument("session_name", type=str, metavar="NAME", nargs=1)
@click.argument("app_name", type=str, metavar="APP", nargs=-1)
@click.option("-l", "--list-names", is_flag=True, help="Just print all available services.")
def apps(session_name, app_name, list_names):
    """
    List available additional arguments and environment variables when starting service.

    \b
    SESSID: The compute session ID.
    APP: The name of service provided by the given session. Repeatable.
         If none provided, this will print all available services.
    """

    async def print_arguments():
        apps = []
        async with AsyncSession() as api_session:
            compute_session = api_session.ComputeSession(session_name)
            apps = await compute_session.stream_app_info()
            if len(app_name) > 0:
                apps = list(filter(lambda x: x["name"] in app_name))
        if list_names:
            print_info(
                "This session provides the following app services: {0}".format(
                    ", ".join(list(map(lambda x: x["name"], apps)))
                )
            )
            return
        for service in apps:
            has_arguments = "allowed_arguments" in service.keys()
            has_envs = "allowed_envs" in service.keys()

            if has_arguments or has_envs:
                print_info("Information for service {0}:".format(service["name"]))
                if has_arguments:
                    print("\tAvailable arguments: {0}".format(service["allowed_arguments"]))
                if has_envs:
                    print("\tAvailable environment variables: {0}".format(service["allowed_envs"]))
            else:
                print_info(
                    "Service {0} does not have customizable arguments.".format(service["name"])
                )

    try:
        asyncio_run(print_arguments())
    except Exception as e:
        print_error(e)



import json
import sys
import time
from pathlib import Path

import appdirs
import click
from tabulate import tabulate

from ai.backend.cli.interaction import ask_yn

from ..pretty import print_done, print_error, print_fail, print_info, print_wait
from ..session import Session
from . import admin


@admin.group()
def manager():
    """Set of manager control operations."""


@manager.command()
def status():
    """Show the manager's current status."""
    try:
        with Session() as session:
            resp = session.Manager.status()
            print(
                tabulate(
                    [("Status", "Active Sessions"), (resp["status"], resp["active_sessions"])],
                    headers="firstrow",
                )
            )
    except Exception as e:
        print_error(e)
        sys.exit(1)


@manager.command()
@click.option(
    "--wait",
    is_flag=True,
    help="Hold up freezing the manager until " "there are no running sessions in the manager.",
)
@click.option(
    "--force-kill",
    is_flag=True,
    help="Kill all running sessions immediately and freeze the manager.",
)
def freeze(wait, force_kill):
    """Freeze manager."""
    if wait and force_kill:
        print(
            "You cannot use both --wait and --force-kill options " "at the same time.",
            file=sys.stderr,
        )
        return
    try:
        with Session() as session:
            if wait:
                while True:
                    resp = session.Manager.status()
                    active_sessions_num = resp["active_sessions"]
                    if active_sessions_num == 0:
                        break
                    print_wait(
                        "Waiting for all sessions terminated... ({0} left)".format(
                            active_sessions_num
                        )
                    )
                    time.sleep(3)
                print_done("All sessions are terminated.")

            if force_kill:
                print_wait("Killing all sessions...")

            session.Manager.freeze(force_kill=force_kill)

            if force_kill:
                print_done("All sessions are killed.")

            print("Manager is successfully frozen.")
    except Exception as e:
        print_error(e)
        sys.exit(1)


@manager.command()
def unfreeze():
    """Unfreeze manager."""
    try:
        with Session() as session:
            session.Manager.unfreeze()
            print("Manager is successfully unfrozen.")
    except Exception as e:
        print_error(e)
        sys.exit(1)


@admin.group()
def announcement():
    """Global announcement related commands"""


@announcement.command()
def get():
    """Get current announcement."""
    try:
        with Session() as session:
            result = session.Manager.get_announcement()
            if result.get("enabled", False):
                msg = result.get("message")
                print(msg)
            else:
                print("No announcements.")
    except Exception as e:
        print_error(e)
        sys.exit(1)


@announcement.command()
@click.option("-m", "--message", default=None, type=click.STRING)
def update(message):
    """
    Post new announcement.

    MESSAGE: Announcement message.
    """
    try:
        with Session() as session:
            if message is None:
                message = click.edit(
                    "<!-- Use Markdown format to edit the announcement message -->",
                )
            if message is None:
                print_info("Cancelled")
                sys.exit(1)
            session.Manager.update_announcement(enabled=True, message=message)
        print_done("Posted new announcement.")
    except Exception as e:
        print_error(e)
        sys.exit(1)


@announcement.command()
def delete():
    """Delete current announcement."""
    if not ask_yn():
        print_info("Cancelled.")
        sys.exit(1)
    try:
        with Session() as session:
            session.Manager.update_announcement(enabled=False)
        print_done("Deleted announcement.")
    except Exception as e:
        print_error(e)
        sys.exit(1)


@announcement.command()
def dismiss():
    """Do not show the same announcement again."""
    if not ask_yn():
        print_info("Cancelled.")
        sys.exit(1)
    try:
        local_state_path = Path(appdirs.user_state_dir("backend.ai", "Lablup"))
        with open(local_state_path / "announcement.json", "rb") as f:
            state = json.load(f)
        state["dismissed"] = True
        with open(local_state_path / "announcement.json", "w") as f:
            json.dump(state, f)
        print_done("Dismissed the last shown announcement.")
    except (IOError, json.JSONDecodeError):
        print_fail("No announcements seen yet.")
        sys.exit(1)
    except Exception as e:
        print_error(e)
        sys.exit(1)


@manager.group()
def scheduler():
    """
    The scheduler operation command group.
    """
    pass


@scheduler.command()
@click.argument("agent_ids", nargs=-1)
def include_agents(agent_ids):
    """
    Include agents in scheduling, meaning that the given agents
    will be considered to be ready for creating new session containers.
    """
    try:
        with Session() as session:
            session.Manager.scheduler_op("include-agents", agent_ids)
        print_done("The given agents now accepts new sessions.")
    except Exception as e:
        print_error(e)
        sys.exit(1)


@scheduler.command()
@click.argument("agent_ids", nargs=-1)
def exclude_agents(agent_ids):
    """
    Exclude agents from scheduling, meaning that the given agents
    will no longer start new sessions unless they are "included" again,
    regardless of their restarts and rejoining events.
    """
    try:
        with Session() as session:
            session.Manager.scheduler_op("exclude-agents", agent_ids)
        print_done("The given agents will no longer start new sessions.")
    except Exception as e:
        print_error(e)
        sys.exit(1)



import asyncio
import collections
import getpass
import itertools
import json
import secrets
import string
import sys
import traceback
from decimal import Decimal
from typing import Mapping, Optional, Sequence, Tuple

import aiohttp
import click
import tabulate as tabulate_mod
from humanize import naturalsize
from tabulate import tabulate

from ai.backend.cli.main import main

from ..compat import asyncio_run, current_loop
from ..config import local_cache_path
from ..exceptions import BackendError
from ..session import AsyncSession
from .params import CommaSeparatedListType, RangeExprOptionType
from .pretty import (
    format_info,
    print_done,
    print_error,
    print_fail,
    print_info,
    print_wait,
    print_warn,
)

tabulate_mod.PRESERVE_WHITESPACE = True
range_expr = RangeExprOptionType()
list_expr = CommaSeparatedListType()


async def exec_loop(
    stdout,
    stderr,
    compute_session,
    mode,
    code,
    *,
    opts=None,
    vprint_done=print_done,
    is_multi=False,
):
    """
    Fully streamed asynchronous version of the execute loop.
    """
    async with compute_session.stream_execute(code, mode=mode, opts=opts) as stream:
        async for result in stream:
            if result.type == aiohttp.WSMsgType.TEXT:
                result = json.loads(result.data)
            else:
                # future extension
                continue
            for rec in result.get("console", []):
                if rec[0] == "stdout":
                    print(rec[1], end="", file=stdout)
                elif rec[0] == "stderr":
                    print(rec[1], end="", file=stderr)
                else:
                    print("----- output record (type: {0}) -----".format(rec[0]), file=stdout)
                    print(rec[1], file=stdout)
                    print("----- end of record -----", file=stdout)
            stdout.flush()
            files = result.get("files", [])
            if files:
                print("--- generated files ---", file=stdout)
                for item in files:
                    print("{0}: {1}".format(item["name"], item["url"]), file=stdout)
                print("--- end of generated files ---", file=stdout)
            if result["status"] == "clean-finished":
                exitCode = result.get("exitCode")
                msg = "Clean finished. (exit code = {0})".format(exitCode)
                if is_multi:
                    print(msg, file=stderr)
                vprint_done(msg)
            elif result["status"] == "build-finished":
                exitCode = result.get("exitCode")
                msg = "Build finished. (exit code = {0})".format(exitCode)
                if is_multi:
                    print(msg, file=stderr)
                vprint_done(msg)
            elif result["status"] == "finished":
                exitCode = result.get("exitCode")
                msg = "Execution finished. (exit code = {0})".format(exitCode)
                if is_multi:
                    print(msg, file=stderr)
                vprint_done(msg)
                break
            elif result["status"] == "waiting-input":
                if result["options"].get("is_password", False):
                    code = getpass.getpass()
                else:
                    code = input()
                await stream.send_str(code)
            elif result["status"] == "continued":
                pass


def exec_loop_sync(
    stdout, stderr, compute_session, mode, code, *, opts=None, vprint_done=print_done
):
    """
    Old synchronous polling version of the execute loop.
    """
    opts = opts if opts else {}
    run_id = None  # use server-assigned run ID
    while True:
        result = compute_session.execute(run_id, code, mode=mode, opts=opts)
        run_id = result["runId"]
        opts.clear()  # used only once
        for rec in result["console"]:
            if rec[0] == "stdout":
                print(rec[1], end="", file=stdout)
            elif rec[0] == "stderr":
                print(rec[1], end="", file=stderr)
            else:
                print("----- output record (type: {0}) -----".format(rec[0]), file=stdout)
                print(rec[1], file=stdout)
                print("----- end of record -----", file=stdout)
        stdout.flush()
        files = result.get("files", [])
        if files:
            print("--- generated files ---", file=stdout)
            for item in files:
                print("{0}: {1}".format(item["name"], item["url"]), file=stdout)
            print("--- end of generated files ---", file=stdout)
        if result["status"] == "clean-finished":
            exitCode = result.get("exitCode")
            vprint_done("Clean finished. (exit code = {0}".format(exitCode), file=stdout)
            mode = "continue"
            code = ""
        elif result["status"] == "build-finished":
            exitCode = result.get("exitCode")
            vprint_done("Build finished. (exit code = {0})".format(exitCode), file=stdout)
            mode = "continue"
            code = ""
        elif result["status"] == "finished":
            exitCode = result.get("exitCode")
            vprint_done("Execution finished. (exit code = {0})".format(exitCode), file=stdout)
            break
        elif result["status"] == "waiting-input":
            mode = "input"
            if result["options"].get("is_password", False):
                code = getpass.getpass()
            else:
                code = input()
        elif result["status"] == "continued":
            mode = "continue"
            code = ""


async def exec_terminal(compute_session, *, vprint_wait=print_wait, vprint_done=print_done):
    # async with compute_session.stream_pty() as stream: ...
    raise NotImplementedError


def _noop(*args, **kwargs):
    pass


def format_stats(stats):
    formatted = []
    version = stats.pop("version", 1)
    stats.pop("status")
    if version == 1:
        stats.pop("precpu_used", None)
        stats.pop("precpu_system_used", None)
        stats.pop("cpu_system_used", None)
        for key, val in stats.items():
            if key.endswith("_size") or key.endswith("_bytes"):
                val = naturalsize(val, binary=True)
            elif key == "cpu_used":
                key += "_msec"
                val = "{0:,}".format(int(val))
            else:
                val = "{0:,}".format(int(val))
            formatted.append((key, val))
    elif version == 2:
        max_integer_len = 0
        max_fraction_len = 0
        for key, metric in stats.items():
            unit = metric["unit_hint"]
            if unit == "bytes":
                val = metric.get("stats.max", metric["current"])
                val = naturalsize(val, binary=True)
                val, unit = val.rsplit(" ", maxsplit=1)
                val = "{:,}".format(Decimal(val))
            elif unit == "msec":
                val = "{:,}".format(Decimal(metric["current"]))
                unit = "msec"
            elif unit == "percent":
                val = metric["pct"]
                unit = "%"
            else:
                val = metric["current"]
                unit = ""
            if val is None:
                continue
            ip, _, fp = val.partition(".")
            max_integer_len = max(len(ip), max_integer_len)
            max_fraction_len = max(len(fp), max_fraction_len)
            formatted.append([key, val, unit])
        fstr_int_only = "{0:>" + str(max_integer_len) + "}"
        fstr_float = "{0:>" + str(max_integer_len) + "}.{1:<" + str(max_fraction_len) + "}"
        for item in formatted:
            ip, _, fp = item[1].partition(".")
            if fp == "":
                item[1] = fstr_int_only.format(ip) + " " * (max_fraction_len + 1)
            else:
                item[1] = fstr_float.format(ip, fp)
    else:
        print_warn("Unsupported statistics result version. Upgrade your client.")
    return tabulate(formatted)


def prepare_resource_arg(resources):
    if resources:
        resources = {k: v for k, v in map(lambda s: s.split("=", 1), resources)}
    else:
        resources = {}  # use the defaults configured in the server
    return resources


def prepare_env_arg(env):
    if env is not None:
        envs = {k: v for k, v in map(lambda s: s.split("=", 1), env)}
    else:
        envs = {}
    return envs


def prepare_mount_arg(
    mount_args: Optional[Sequence[str]],
) -> Tuple[Sequence[str], Mapping[str, str]]:
    """
    Parse the list of mount arguments into a list of
    vfolder name and in-container mount path pairs.
    """
    mounts = set()
    mount_map = {}
    if mount_args is not None:
        for value in mount_args:
            if "=" in value:
                sp = value.split("=", maxsplit=1)
            elif ":" in value:  # docker-like volume mount mapping
                sp = value.split(":", maxsplit=1)
            else:
                sp = [value]
            mounts.add(sp[0])
            if len(sp) == 2:
                mount_map[sp[0]] = sp[1]
    return list(mounts), mount_map


@main.command()
@click.argument("image", type=str)
@click.argument("files", nargs=-1, type=click.Path())
@click.option(
    "-t",
    "--name",
    "--client-token",
    metavar="NAME",
    help="Specify a human-readable session name. " "If not set, a random hex string is used.",
)
# job scheduling options
@click.option(
    "--type",
    metavar="SESSTYPE",
    type=click.Choice(["batch", "interactive"]),
    default="interactive",
    help="Either batch or interactive",
)
@click.option(
    "--starts-at",
    metavar="STARTS_AT",
    type=str,
    default=None,
    help="Let session to be started at a specific or relative time.",
)
@click.option(
    "--enqueue-only",
    is_flag=True,
    help="Enqueue the session and return immediately without waiting for its startup.",
)
@click.option(
    "--max-wait",
    metavar="SECONDS",
    type=int,
    default=0,
    help="The maximum duration to wait until the session starts.",
)
@click.option(
    "--no-reuse", is_flag=True, help="Do not reuse existing sessions but return an error."
)
@click.option(
    "--callback-url",
    metavar="CALLBACK_URL",
    type=str,
    default=None,
    help="Callback URL which will be called upon sesison lifecycle events.",
)
# query-mode options
@click.option("-c", "--code", metavar="CODE", help="The code snippet as a single string")
@click.option("--terminal", is_flag=True, help="Connect to the terminal-type compute_session.")
# batch-mode options
@click.option(
    "--clean", metavar="CMD", help="Custom shell command for cleaning up the base directory"
)
@click.option("--build", metavar="CMD", help="Custom shell command for building the given files")
@click.option("--exec", metavar="CMD", help="Custom shell command for executing the given files")
@click.option(
    "--basedir",
    metavar="PATH",
    type=click.Path(),
    default=None,
    help="Base directory path of uploaded files. "
    "All uploaded files must reside inside this directory.",
)
# execution environment
@click.option(
    "-e",
    "--env",
    metavar="KEY=VAL",
    type=str,
    multiple=True,
    help="Environment variable (may appear multiple times)",
)
# extra options
@click.option(
    "--bootstrap-script",
    metavar="PATH",
    type=click.File("r"),
    default=None,
    help="A user-defined script to execute on startup.",
)
@click.option(
    "--rm",
    is_flag=True,
    help="Terminate the session immediately after running " "the given code or files",
)
@click.option(
    "-s",
    "--stats",
    is_flag=True,
    help="Show resource usage statistics after termination " '(only works if "--rm" is given)',
)
@click.option("--tag", type=str, default=None, help="User-defined tag string to annotate sessions.")
@click.option(
    "-q",
    "--quiet",
    is_flag=True,
    help="Hide execution details but show only the compute_session outputs.",
)
# experiment support
@click.option(
    "--env-range",
    metavar="RANGE_EXPR",
    multiple=True,
    type=range_expr,
    help="Range expression for environment variable.",
)
@click.option(
    "--build-range",
    metavar="RANGE_EXPR",
    multiple=True,
    type=range_expr,
    help="Range expression for execution arguments.",
)
@click.option(
    "--exec-range",
    metavar="RANGE_EXPR",
    multiple=True,
    type=range_expr,
    help="Range expression for execution arguments.",
)
@click.option(
    "--max-parallel",
    metavar="NUM",
    type=int,
    default=2,
    help="The maximum number of parallel sessions.",
)
# resource spec
@click.option(
    "-v",
    "--volume",
    "-m",
    "--mount",
    "mount",
    metavar="NAME[=PATH]",
    type=str,
    multiple=True,
    help="User-owned virtual folder names to mount. "
    "If path is not provided, virtual folder will be mounted under /home/work. "
    "When the target path is relative, it is placed under /home/work "
    "with auto-created parent directories if any. "
    "Absolute paths are mounted as-is, but it is prohibited to "
    "override the predefined Linux system directories.",
)
@click.option(
    "--scaling-group",
    "--sgroup",
    type=str,
    default=None,
    help="The scaling group to execute session. If not specified, "
    "all available scaling groups are included in the scheduling.",
)
@click.option(
    "-r",
    "--resources",
    "--resource",
    metavar="KEY=VAL",
    type=str,
    multiple=True,
    help="Set computation resources " "(e.g: -r cpu=2 -r mem=256 -r cuda.device=1)",
)
@click.option(
    "--cluster-size",
    metavar="NUMBER",
    type=int,
    default=1,
    help="The size of cluster in number of containers.",
)
@click.option(
    "--cluster-mode",
    metavar="MODE",
    type=click.Choice(["single-node", "multi-node"]),
    default="single-node",
    help="The mode of clustering.",
)
@click.option(
    "--resource-opts",
    metavar="KEY=VAL",
    type=str,
    multiple=True,
    help="Resource options for creating compute session. " "(e.g: shmem=64m)",
)
# resource grouping
@click.option(
    "-d",
    "--domain",
    metavar="DOMAIN_NAME",
    default=None,
    help="Domain name where the session will be spawned. "
    "If not specified, config's domain name will be used.",
)
@click.option(
    "-g",
    "--group",
    metavar="GROUP_NAME",
    default=None,
    help="Group name where the session is spawned. "
    "User should be a member of the group to execute the code.",
)
@click.option("--preopen", default=None, type=list_expr, help="Pre-open service ports")
@click.option(
    "--assign-agent",
    default=None,
    type=list_expr,
    help="Show mapping list of tuple which mapped containers with agent. "
    "When user role is Super Admin. "
    "(e.g., --assign-agent agent_id_1,agent_id_2,...)",
)
def run(
    image,
    files,
    name,  # base args
    type,
    starts_at,
    enqueue_only,
    max_wait,
    no_reuse,  # job scheduling options
    callback_url,
    code,
    terminal,  # query-mode options
    clean,
    build,
    exec,
    basedir,  # batch-mode options
    env,  # execution environment
    bootstrap_script,
    rm,
    stats,
    tag,
    quiet,  # extra options
    env_range,
    build_range,
    exec_range,
    max_parallel,  # experiment support
    mount,
    scaling_group,
    resources,  # resource spec
    cluster_size,
    cluster_mode,
    resource_opts,
    domain,
    group,
    preopen,
    assign_agent,  # resource grouping
):
    """
    Run the given code snippet or files in a session.
    Depending on the session ID you give (default is random),
    it may reuse an existing session or create a new one.

    \b
    IMAGE: The name (and version/platform tags appended after a colon) of session
          runtime or programming language.')
    FILES: The code file(s). Can be added multiple times.
    """
    if quiet:
        vprint_info = vprint_wait = vprint_done = _noop
    else:
        vprint_info = print_info
        vprint_wait = print_wait
        vprint_done = print_done
    if files and code:
        print(
            "You can run only either source files or command-line " "code snippet.", file=sys.stderr
        )
        sys.exit(1)
    if not files and not code:
        print(
            "You should provide the command-line code snippet using "
            '"-c" option if run without files.',
            file=sys.stderr,
        )
        sys.exit(1)

    envs = prepare_env_arg(env)
    resources = prepare_resource_arg(resources)
    resource_opts = prepare_resource_arg(resource_opts)
    mount, mount_map = prepare_mount_arg(mount)

    if env_range is None:
        env_range = []  # noqa
    if build_range is None:
        build_range = []  # noqa
    if exec_range is None:
        exec_range = []  # noqa

    env_ranges = {v: r for v, r in env_range}
    build_ranges = {v: r for v, r in build_range}
    exec_ranges = {v: r for v, r in exec_range}

    env_var_maps = [
        dict(zip(env_ranges.keys(), values)) for values in itertools.product(*env_ranges.values())
    ]
    build_var_maps = [
        dict(zip(build_ranges.keys(), values))
        for values in itertools.product(*build_ranges.values())
    ]
    exec_var_maps = [
        dict(zip(exec_ranges.keys(), values)) for values in itertools.product(*exec_ranges.values())
    ]
    case_set = collections.OrderedDict()
    vmaps_product = itertools.product(env_var_maps, build_var_maps, exec_var_maps)
    build_template = string.Template(build)
    exec_template = string.Template(exec)
    env_templates = {k: string.Template(v) for k, v in envs.items()}

    if preopen is None:
        preopen = []  # noqa

    preopen_ports = preopen
    assigned_agent_list = assign_agent  # should be None if not specified
    for env_vmap, build_vmap, exec_vmap in vmaps_product:
        interpolated_envs = tuple((k, vt.substitute(env_vmap)) for k, vt in env_templates.items())
        if build:
            interpolated_build = build_template.substitute(build_vmap)
        else:
            interpolated_build = "*"
        if exec:
            interpolated_exec = exec_template.substitute(exec_vmap)
        else:
            interpolated_exec = "*"
        case_set[(interpolated_envs, interpolated_build, interpolated_exec)] = 1

    is_multi = len(case_set) > 1
    if is_multi:
        if max_parallel <= 0:
            print(
                "The number maximum parallel sessions must be " "a positive integer.",
                file=sys.stderr,
            )
            sys.exit(1)
        if terminal:
            print("You cannot run multiple cases with terminal.", file=sys.stderr)
            sys.exit(1)
        if not quiet:
            vprint_info("Running multiple sessions for the following combinations:")
            for case in case_set.keys():
                pretty_env = " ".join("{}={}".format(item[0], item[1]) for item in case[0])
                print("env = {!r}, build = {!r}, exec = {!r}".format(pretty_env, case[1], case[2]))

    def _run_legacy(session, idx, name, envs, clean_cmd, build_cmd, exec_cmd):
        try:
            compute_session = session.ComputeSession.get_or_create(
                image,
                name=name,
                type_=type,
                enqueue_only=enqueue_only,
                max_wait=max_wait,
                no_reuse=no_reuse,
                cluster_size=cluster_size,
                cluster_mode=cluster_mode,
                mounts=mount,
                mount_map=mount_map,
                envs=envs,
                resources=resources,
                domain_name=domain,
                group_name=group,
                scaling_group=scaling_group,
                tag=tag,
            )
        except Exception as e:
            print_error(e)
            sys.exit(1)
        if compute_session.status == "PENDING":
            print_info("Session ID {0} is enqueued for scheduling.".format(name))
            return
        elif compute_session.status == "SCHEDULED":
            print_info("Session ID {0} is scheduled and about to be started.".format(name))
            return
        elif compute_session.status == "RUNNING":
            if compute_session.created:
                vprint_done(
                    "[{0}] Session {1} is ready (domain={2}, group={3}).".format(
                        idx, compute_session.name, compute_session.domain, compute_session.group
                    )
                )
            else:
                vprint_done("[{0}] Reusing session {1}...".format(idx, compute_session.name))
        elif compute_session.status == "TERMINATED":
            print_warn(
                "Session ID {0} is already terminated.\n"
                "This may be an error in the compute_session image.".format(name)
            )
            return
        elif compute_session.status == "TIMEOUT":
            print_info("Session ID {0} is still on the job queue.".format(name))
            return
        elif compute_session.status in ("ERROR", "CANCELLED"):
            print_fail(
                "Session ID {0} has an error during scheduling/startup or cancelled.".format(name)
            )
            return

        try:
            if files:
                vprint_wait("[{0}] Uploading source files...".format(idx))
                ret = compute_session.upload(files, basedir=basedir, show_progress=True)
                if ret.status // 100 != 2:
                    print_fail("[{0}] Uploading source files failed!".format(idx))
                    print("{0}: {1}\n{2}".format(ret.status, ret.reason, ret.text()))
                    return
                vprint_done("[{0}] Uploading done.".format(idx))
                opts = {
                    "clean": clean_cmd,
                    "build": build_cmd,
                    "exec": exec_cmd,
                }
                if not terminal:
                    exec_loop_sync(
                        sys.stdout,
                        sys.stderr,
                        compute_session,
                        "batch",
                        "",
                        opts=opts,
                        vprint_done=vprint_done,
                    )
            if terminal:
                raise NotImplementedError(
                    "Terminal access is not supported in " "the legacy synchronous mode."
                )
            if code:
                exec_loop_sync(
                    sys.stdout, sys.stderr, compute_session, "query", code, vprint_done=vprint_done
                )
            vprint_done("[{0}] Execution finished.".format(idx))
        except Exception as e:
            print_error(e)
            sys.exit(1)
        finally:
            if rm:
                vprint_wait("[{0}] Cleaning up the session...".format(idx))
                ret = compute_session.destroy()
                vprint_done("[{0}] Cleaned up the session.".format(idx))
                if stats:
                    _stats = ret.get("stats", None) if ret else None
                    if _stats:
                        print("[{0}] Statistics:\n{1}".format(idx, format_stats(_stats)))
                    else:
                        print("[{0}] Statistics is not available.".format(idx))

    async def _run(session, idx, name, envs, clean_cmd, build_cmd, exec_cmd, is_multi=False):
        try:
            compute_session = await session.ComputeSession.get_or_create(
                image,
                name=name,
                type_=type,
                starts_at=starts_at,
                enqueue_only=enqueue_only,
                max_wait=max_wait,
                no_reuse=no_reuse,
                callback_url=callback_url,
                cluster_size=cluster_size,
                cluster_mode=cluster_mode,
                mounts=mount,
                mount_map=mount_map,
                envs=envs,
                resources=resources,
                resource_opts=resource_opts,
                domain_name=domain,
                group_name=group,
                scaling_group=scaling_group,
                bootstrap_script=bootstrap_script.read() if bootstrap_script is not None else None,
                tag=tag,
                preopen_ports=preopen_ports,
                assign_agent=assigned_agent_list,
            )
        except Exception as e:
            print_fail("[{0}] {1}".format(idx, e))
            return
        if compute_session.status == "PENDING":
            print_info("Session ID {0} is enqueued for scheduling.".format(name))
            return
        elif compute_session.status == "SCHEDULED":
            print_info("Session ID {0} is scheduled and about to be started.".format(name))
            return
        elif compute_session.status == "RUNNING":
            if compute_session.created:
                vprint_done(
                    "[{0}] Session {1} is ready (domain={2}, group={3}).".format(
                        idx, compute_session.name, compute_session.domain, compute_session.group
                    )
                )
            else:
                vprint_done("[{0}] Reusing session {1}...".format(idx, compute_session.name))
        elif compute_session.status == "TERMINATED":
            print_warn(
                "Session ID {0} is already terminated.\n"
                "This may be an error in the compute_session image.".format(name)
            )
            return
        elif compute_session.status == "TIMEOUT":
            print_info("Session ID {0} is still on the job queue.".format(name))
            return
        elif compute_session.status in ("ERROR", "CANCELLED"):
            print_fail(
                "Session ID {0} has an error during scheduling/startup or cancelled.".format(name)
            )
            return

        if not is_multi:
            stdout = sys.stdout
            stderr = sys.stderr
        else:
            log_dir = local_cache_path / "client-logs"
            log_dir.mkdir(parents=True, exist_ok=True)
            stdout = open(log_dir / "{0}.stdout.log".format(name), "w", encoding="utf-8")
            stderr = open(log_dir / "{0}.stderr.log".format(name), "w", encoding="utf-8")

        try:

            def indexed_vprint_done(msg):
                vprint_done("[{0}] ".format(idx) + msg)

            if files:
                if not is_multi:
                    vprint_wait("[{0}] Uploading source files...".format(idx))
                ret = await compute_session.upload(
                    files, basedir=basedir, show_progress=not is_multi
                )
                if ret.status // 100 != 2:
                    print_fail("[{0}] Uploading source files failed!".format(idx))
                    print("{0}: {1}\n{2}".format(ret.status, ret.reason, ret.text()), file=stderr)
                    raise RuntimeError("Uploading source files has failed!")
                if not is_multi:
                    vprint_done("[{0}] Uploading done.".format(idx))
                opts = {
                    "clean": clean_cmd,
                    "build": build_cmd,
                    "exec": exec_cmd,
                }
                if not terminal:
                    await exec_loop(
                        stdout,
                        stderr,
                        compute_session,
                        "batch",
                        "",
                        opts=opts,
                        vprint_done=indexed_vprint_done,
                        is_multi=is_multi,
                    )
            if terminal:
                await exec_terminal(compute_session)
                return
            if code:
                await exec_loop(
                    stdout,
                    stderr,
                    compute_session,
                    "query",
                    code,
                    vprint_done=indexed_vprint_done,
                    is_multi=is_multi,
                )
        except BackendError as e:
            print_fail("[{0}] {1}".format(idx, e))
            raise RuntimeError(e)
        except Exception as e:
            print_fail("[{0}] Execution failed!".format(idx))
            traceback.print_exc()
            raise RuntimeError(e)
        finally:
            try:
                if rm:
                    if not is_multi:
                        vprint_wait("[{0}] Cleaning up the session...".format(idx))
                    ret = await compute_session.destroy()
                    vprint_done("[{0}] Cleaned up the session.".format(idx))
                    if stats:
                        _stats = ret.get("stats", None) if ret else None
                        if _stats:
                            stats_str = format_stats(_stats)
                            print(
                                format_info("[{0}] Statistics:".format(idx))
                                + "\n{0}".format(stats_str)
                            )
                            if is_multi:
                                print("Statistics:\n{0}".format(stats_str), file=stderr)
                        else:
                            print_warn("[{0}] Statistics: unavailable.".format(idx))
                            if is_multi:
                                print("Statistics: unavailable.", file=stderr)
            except Exception as e:
                print_fail("[{0}] Error while printing stats".format(idx))
                traceback.print_exc()
                raise RuntimeError(e)
            finally:
                if is_multi:
                    stdout.close()
                    stderr.close()

    async def _run_cases():
        loop = current_loop()
        if name is None:
            name_prefix = f"pysdk-{secrets.token_hex(5)}"
        else:
            name_prefix = name
        vprint_info("Session name prefix: {0}".format(name_prefix))
        if is_multi:
            print_info(
                "Check out the stdout/stderr logs stored in "
                "~/.cache/backend.ai/client-logs directory."
            )
        async with AsyncSession() as session:
            tasks = []
            # TODO: limit max-parallelism using aiojobs
            for idx, case in enumerate(case_set.keys()):
                if is_multi:
                    _name = "{0}-{1}".format(name_prefix, idx)
                else:
                    _name = name_prefix
                envs = dict(case[0])
                clean_cmd = clean if clean else "*"
                build_cmd = case[1]
                exec_cmd = case[2]
                t = loop.create_task(
                    _run(
                        session, idx, _name, envs, clean_cmd, build_cmd, exec_cmd, is_multi=is_multi
                    )
                )
                tasks.append(t)
            results = await asyncio.gather(*tasks, return_exceptions=True)
            if any(map(lambda r: isinstance(r, Exception), results)):
                if is_multi:
                    print_fail("There were failed cases!")
                sys.exit(1)

    try:
        asyncio_run(_run_cases())
    except Exception as e:
        print_fail("{0}".format(e))



import textwrap
from typing import Iterable, Sequence

from ai.backend.client.output.fields import group_fields
from ai.backend.client.output.types import FieldSpec

from ..session import api_session
from .base import BaseFunction, api_function, resolve_fields

__all__ = ("Group",)

_default_list_fields = (
    group_fields["id"],
    group_fields["name"],
    group_fields["is_active"],
    group_fields["created_at"],
    group_fields["integration_id"],
)
_default_detail_fields = (
    group_fields["id"],
    group_fields["name"],
    group_fields["description"],
    group_fields["is_active"],
    group_fields["created_at"],
    group_fields["domain_name"],
    group_fields["total_resource_slots"],
    group_fields["allowed_vfolder_hosts"],
    group_fields["integration_id"],
)


class Group(BaseFunction):
    """
    Provides a shortcut of :func:`Group.query()
    <ai.backend.client.admin.Admin.query>` that fetches various group information.

    .. note::

      All methods in this function class require your API access key to
      have the *admin* privilege.
    """

    @api_function
    @classmethod
    async def from_name(
        cls,
        name: str,
        *,
        fields: Iterable[FieldSpec | str] = None,
        domain_name: str = None,
    ) -> Sequence[dict]:
        """
        Find the group(s) by its name.
        It may return multiple groups when there are groups with the same name
        in different domains and it is invoked with a super-admin account
        without setting the domain name.

        :param domain_name: Name of domain to get groups from.
        :param fields: Per-group query fields to fetch.
        """
        query = textwrap.dedent(
            """\
            query($name: String!, $domain_name: String) {
                groups_by_name(name: $name, domain_name: $domain_name) {$fields}
            }
        """
        )
        resolved_fields = resolve_fields(fields, group_fields, _default_detail_fields)
        query = query.replace("$fields", " ".join(resolved_fields))
        variables = {
            "name": name,
            "domain_name": domain_name,
        }
        data = await api_session.get().Admin._query(query, variables)
        return data["groups_by_name"]

    @api_function
    @classmethod
    async def list(
        cls,
        domain_name: str,
        fields: Sequence[FieldSpec] = _default_list_fields,
    ) -> Sequence[dict]:
        """
        Fetches the list of groups.

        :param domain_name: Name of domain to list groups.
        :param fields: Per-group query fields to fetch.
        """
        if fields is None:
            fields = _default_list_fields
        query = textwrap.dedent(
            """\
            query($domain_name: String) {
                groups(domain_name: $domain_name) {$fields}
            }
        """
        )
        query = query.replace("$fields", " ".join(f.field_ref for f in fields))
        variables = {"domain_name": domain_name}
        data = await api_session.get().Admin._query(query, variables)
        return data["groups"]

    @api_function
    @classmethod
    async def detail(
        cls,
        gid: str,
        fields: Sequence[FieldSpec] = _default_detail_fields,
    ) -> dict:
        """
        Fetch information of a group with group ID.

        :param gid: ID of the group to fetch.
        :param fields: Additional per-group query fields to fetch.
        """
        if fields is None:
            fields = _default_detail_fields
        query = textwrap.dedent(
            """\
            query($gid: UUID!) {
                group(id: $gid) {$fields}
            }
        """
        )
        query = query.replace("$fields", " ".join(f.field_ref for f in fields))
        variables = {"gid": gid}
        data = await api_session.get().Admin._query(query, variables)
        return data["group"]

    @api_function
    @classmethod
    async def create(
        cls,
        domain_name: str,
        name: str,
        description: str = "",
        is_active: bool = True,
        total_resource_slots: str = None,
        allowed_vfolder_hosts: Iterable[str] = None,
        integration_id: str = None,
        fields: Iterable[FieldSpec | str] = None,
    ) -> dict:
        """
        Creates a new group with the given options.
        You need an admin privilege for this operation.
        """
        query = textwrap.dedent(
            """\
            mutation($name: String!, $input: GroupInput!) {
                create_group(name: $name, props: $input) {
                    ok msg group {$fields}
                }
            }
        """
        )
        resolved_fields = resolve_fields(
            fields,
            group_fields,
            (group_fields["id"], group_fields["domain_name"], group_fields["name"]),
        )
        query = query.replace("$fields", " ".join(resolved_fields))
        variables = {
            "name": name,
            "input": {
                "description": description,
                "is_active": is_active,
                "domain_name": domain_name,
                "total_resource_slots": total_resource_slots,
                "allowed_vfolder_hosts": allowed_vfolder_hosts,
                "integration_id": integration_id,
            },
        }
        data = await api_session.get().Admin._query(query, variables)
        return data["create_group"]

    @api_function
    @classmethod
    async def update(
        cls,
        gid: str,
        name: str = None,
        description: str = None,
        is_active: bool = None,
        total_resource_slots: str = None,
        allowed_vfolder_hosts: Iterable[str] = None,
        integration_id: str = None,
        fields: Iterable[FieldSpec | str] = None,
    ) -> dict:
        """
        Update existing group.
        You need an admin privilege for this operation.
        """
        query = textwrap.dedent(
            """\
            mutation($gid: UUID!, $input: ModifyGroupInput!) {
                modify_group(gid: $gid, props: $input) {
                    ok msg
                }
            }
        """
        )
        variables = {
            "gid": gid,
            "input": {
                "name": name,
                "description": description,
                "is_active": is_active,
                "total_resource_slots": total_resource_slots,
                "allowed_vfolder_hosts": allowed_vfolder_hosts,
                "integration_id": integration_id,
            },
        }
        data = await api_session.get().Admin._query(query, variables)
        return data["modify_group"]

    @api_function
    @classmethod
    async def delete(cls, gid: str):
        """
        Inactivates the existing group. Does not actually delete it for safety.
        """
        query = textwrap.dedent(
            """\
            mutation($gid: UUID!) {
                delete_group(gid: $gid) {
                    ok msg
                }
            }
        """
        )
        variables = {"gid": gid}
        data = await api_session.get().Admin._query(query, variables)
        return data["delete_group"]

    @api_function
    @classmethod
    async def purge(cls, gid: str):
        """
        Delete the existing group. This action cannot be undone.
        """
        query = textwrap.dedent(
            """\
            mutation($gid: UUID!) {
                purge_group(gid: $gid) {
                    ok msg
                }
            }
        """
        )
        variables = {"gid": gid}
        data = await api_session.get().Admin._query(query, variables)
        return data["purge_group"]

    @api_function
    @classmethod
    async def add_users(
        cls, gid: str, user_uuids: Iterable[str], fields: Iterable[FieldSpec | str] = None
    ) -> dict:
        """
        Add users to a group.
        You need an admin privilege for this operation.
        """
        query = textwrap.dedent(
            """\
            mutation($gid: UUID!, $input: ModifyGroupInput!) {
                modify_group(gid: $gid, props: $input) {
                    ok msg
                }
            }
        """
        )
        variables = {
            "gid": gid,
            "input": {
                "user_update_mode": "add",
                "user_uuids": user_uuids,
            },
        }
        data = await api_session.get().Admin._query(query, variables)
        return data["modify_group"]

    @api_function
    @classmethod
    async def remove_users(
        cls, gid: str, user_uuids: Iterable[str], fields: Iterable[FieldSpec | str] = None
    ) -> dict:
        """
        Remove users from a group.
        You need an admin privilege for this operation.
        """
        query = textwrap.dedent(
            """\
            mutation($gid: UUID!, $input: ModifyGroupInput!) {
                modify_group(gid: $gid, props: $input) {
                    ok msg
                }
            }
        """
        )
        variables = {
            "gid": gid,
            "input": {
                "user_update_mode": "remove",
                "user_uuids": user_uuids,
            },
        }
        data = await api_session.get().Admin._query(query, variables)
        return data["modify_group"]



from __future__ import annotations

from .formatters import (
    AgentStatFormatter,
    ContainerListFormatter,
    DependencyListFormatter,
    GroupListFormatter,
    KernelStatFormatter,
    SubFieldOutputFormatter,
    mibytes_output_formatter,
    nested_dict_formatter,
    resource_slot_formatter,
    sizebytes_output_formatter,
)
from .types import FieldSet, FieldSpec

container_fields = FieldSet(
    [
        FieldSpec("id", "Kernel ID", alt_name="kernel_id"),
        FieldSpec("cluster_role"),
        FieldSpec("cluster_idx"),
        FieldSpec("cluster_hostname"),
        FieldSpec("session_id", "Session ID"),
        FieldSpec("image"),
        FieldSpec("registry"),
        FieldSpec("status"),
        FieldSpec("status_info"),
        FieldSpec("status_data", formatter=nested_dict_formatter),
        FieldSpec("status_changed"),
        FieldSpec("agent"),
        FieldSpec("container_id"),
        FieldSpec("resource_opts", formatter=nested_dict_formatter),
        FieldSpec("occupied_slots", formatter=resource_slot_formatter),
        FieldSpec("live_stat", formatter=KernelStatFormatter()),
        FieldSpec("last_stat", formatter=KernelStatFormatter()),
    ]
)


agent_fields = FieldSet(
    [
        FieldSpec("id"),
        FieldSpec("status"),
        FieldSpec("status_changed"),
        FieldSpec("region"),
        FieldSpec("architecture"),
        FieldSpec("scaling_group"),
        FieldSpec("schedulable"),
        FieldSpec("available_slots", formatter=resource_slot_formatter),
        FieldSpec("occupied_slots", formatter=resource_slot_formatter),
        FieldSpec("addr"),
        FieldSpec("first_contact"),
        FieldSpec("lost_at"),
        FieldSpec("live_stat", formatter=AgentStatFormatter()),
        FieldSpec("version"),
        FieldSpec("compute_plugins"),
        FieldSpec("hardware_metadata", formatter=nested_dict_formatter),
        FieldSpec(
            "compute_containers", subfields=container_fields, formatter=ContainerListFormatter()
        ),
        # legacy fields
        FieldSpec("cpu_cur_pct", "CPU Usage (%)"),
        FieldSpec("mem_cur_bytes", "Used Memory (MiB)", formatter=mibytes_output_formatter),
    ]
)

domain_fields = FieldSet(
    [
        FieldSpec("name"),
        FieldSpec("description"),
        FieldSpec("is_active"),
        FieldSpec("created_at"),
        FieldSpec("total_resource_slots", formatter=resource_slot_formatter),
        FieldSpec("allowed_vfolder_hosts"),
        FieldSpec("allowed_docker_registries"),
        FieldSpec("integration_id"),
    ]
)

group_fields = FieldSet(
    [
        FieldSpec("id"),
        FieldSpec("name"),
        FieldSpec("description"),
        FieldSpec("is_active"),
        FieldSpec("created_at"),
        FieldSpec("domain_name"),
        FieldSpec("total_resource_slots", formatter=resource_slot_formatter),
        FieldSpec("allowed_vfolder_hosts"),
        FieldSpec("integration_id"),
    ]
)


image_fields = FieldSet(
    [
        FieldSpec("name"),
        FieldSpec("registry"),
        FieldSpec("architecture"),
        FieldSpec("tag"),
        FieldSpec("digest"),
        FieldSpec("size_bytes", formatter=sizebytes_output_formatter),
        FieldSpec("aliases"),
    ]
)


keypair_fields = FieldSet(
    [
        FieldSpec("user_id", "Email"),
        FieldSpec(
            "user_info { full_name }",
            "Full Name",
            alt_name="full_name",
            formatter=SubFieldOutputFormatter("full_name"),
        ),
        FieldSpec("access_key"),
        FieldSpec("secret_key"),
        FieldSpec("is_active"),
        FieldSpec("is_admin"),
        FieldSpec("created_at"),
        FieldSpec("modified_at"),
        FieldSpec("last_used"),
        FieldSpec("resource_policy"),
        FieldSpec("rate_limit"),
        FieldSpec("concurrency_used"),
        FieldSpec("ssh_public_key"),
        FieldSpec("ssh_private_key"),
        FieldSpec("dotfiles"),
        FieldSpec("bootstrap_script"),
    ]
)


keypair_resource_policy_fields = FieldSet(
    [
        FieldSpec("name"),
        FieldSpec("created_at"),
        FieldSpec("total_resource_slots"),
        FieldSpec("max_concurrent_sessions"),  # formerly concurrency_limit
        FieldSpec("max_vfolder_count"),
        FieldSpec("max_vfolder_size", formatter=sizebytes_output_formatter),
        FieldSpec("idle_timeout"),
        FieldSpec("max_containers_per_session"),
        FieldSpec("allowed_vfolder_hosts"),
    ]
)


scaling_group_fields = FieldSet(
    [
        FieldSpec("name"),
        FieldSpec("description"),
        FieldSpec("is_active"),
        FieldSpec("created_at"),
        FieldSpec("driver"),
        FieldSpec("driver_opts", formatter=nested_dict_formatter),
        FieldSpec("scheduler"),
        FieldSpec("scheduler_opts", formatter=nested_dict_formatter),
    ]
)


session_fields = FieldSet(
    [
        FieldSpec("id", "Kernel ID", alt_name="kernel_id"),
        FieldSpec("tag"),
        FieldSpec("name"),
        FieldSpec("type"),
        FieldSpec("session_id", "Session ID"),
        FieldSpec("image"),
        FieldSpec("registry"),
        FieldSpec("cluster_template"),
        FieldSpec("cluster_mode"),
        FieldSpec("cluster_size"),
        FieldSpec("domain_name"),
        FieldSpec("group_name", "Project/Group"),
        FieldSpec("group_id"),
        FieldSpec("user_email"),
        FieldSpec("user_id"),
        FieldSpec("access_key", "Owner Access Key"),
        FieldSpec("created_user_email"),
        FieldSpec("created_user_id"),
        FieldSpec("status"),
        FieldSpec("status_info"),
        FieldSpec("status_data", formatter=nested_dict_formatter),
        FieldSpec("status_changed", "Last Updated"),
        FieldSpec("created_at"),
        FieldSpec("terminated_at"),
        FieldSpec("starts_at"),
        FieldSpec("startup_command"),
        FieldSpec("result"),
        FieldSpec("resoucre_opts", formatter=nested_dict_formatter),
        FieldSpec("scaling_group"),
        FieldSpec("service_ports", formatter=nested_dict_formatter),
        FieldSpec("mounts"),
        FieldSpec("occupied_slots", formatter=resource_slot_formatter),
        FieldSpec(
            "containers",
            subfields=container_fields,
            formatter=ContainerListFormatter(),
        ),
        FieldSpec(
            "dependencies { name id }",
            formatter=DependencyListFormatter(),
        ),
    ]
)

session_fields_v5 = FieldSet(
    [
        FieldSpec(
            "containers",
            subfields=FieldSet(
                [
                    FieldSpec("id", "Kernel ID", alt_name="kernel_id"),
                    FieldSpec("session_id", "Session ID"),
                    FieldSpec("role"),
                    FieldSpec("agent"),
                    FieldSpec("image"),
                    FieldSpec("status"),
                    FieldSpec("status_info"),
                    FieldSpec("status_data", formatter=nested_dict_formatter),
                    FieldSpec("status_changed"),
                    FieldSpec("occupied_slots", formatter=resource_slot_formatter),
                    FieldSpec("live_stat", formatter=KernelStatFormatter()),
                    FieldSpec("last_stat", formatter=KernelStatFormatter()),
                ]
            ),
            formatter=ContainerListFormatter(),
        ),
    ]
)


storage_fields = FieldSet(
    [
        FieldSpec("id"),
        FieldSpec("backend"),
        FieldSpec("fsprefix"),
        FieldSpec("path"),
        FieldSpec("capabilities"),
        FieldSpec("hardware_metadata", formatter=nested_dict_formatter),
        FieldSpec("performance_metric", formatter=nested_dict_formatter),
        FieldSpec("usage", formatter=nested_dict_formatter),
    ]
)


user_fields = FieldSet(
    [
        FieldSpec("uuid"),
        FieldSpec("username"),
        FieldSpec("email"),
        # password is not queriable!
        FieldSpec("need_password_change"),
        FieldSpec("full_name"),
        FieldSpec("description"),
        FieldSpec("is_active"),
        FieldSpec("status"),
        FieldSpec("status_info"),
        FieldSpec("created_at"),
        FieldSpec("modified_at"),
        FieldSpec("domain_name"),
        FieldSpec("role"),
        FieldSpec("groups { id name }", formatter=GroupListFormatter()),
    ]
)


vfolder_fields = FieldSet(
    [
        FieldSpec("id"),
        FieldSpec("host"),
        FieldSpec("name"),
        FieldSpec("user", alt_name="user_id"),
        FieldSpec("group", alt_name="group_id"),
        FieldSpec("creator"),
        FieldSpec("unmanaged_path"),
        FieldSpec("usage_mode"),
        FieldSpec("permission"),
        FieldSpec("ownership_type"),
        FieldSpec("max_files"),
        FieldSpec("max_size"),
        FieldSpec("created_at"),
        FieldSpec("last_used"),
        FieldSpec("num_files"),
        FieldSpec("cur_size"),
        FieldSpec("cloneable"),
    ]
)
